/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2009 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: phg.h,v 1.107 2009/05/08 03:14:48 zlb Exp $ */

#ifndef PHG_PHG_H

/*#ifdef HAVE_CONFIG_H */
#include "config.h"
/*#endif*/
#include <stdio.h>
#if USE_MPI
# include <mpi.h>
#endif	/* USE_MPI */
#if USE_OMP
# include <omp.h>
#endif	/* USE_OMP */
#include "expression.h"

#ifndef __GNUC__
# define inline
#endif

#define PHG_VERSION		"0.7.0"		/**< Version string */
#define PHG_VERSION_MAJOR	0		/**< Major version number */
#define PHG_VERSION_MINOR	7		/**< Minor version number */
#define PHG_VERSION_SUBMINOR	0		/**< Subminor version number */

#ifndef Dim
# define Dim		3		/**< The Space dimensions */
#endif

#define NVert		(Dim + 1)	/**< Number of vertices / element */
#define NFace		NVert		/**< Number of faces / element */
#if Dim == 3
# define NEdge		6		/**< Number of edges / element */
#else
# define NEdge		NFace
#endif

/* elementary data types */

#if USE_LONG_DOUBLE
 typedef long double	FLOAT;		/**< real, double or float */
#else	/* USE_LONG_DOUBLE */
 typedef double		FLOAT;		/**< real, double or float */
#endif	/* USE_LONG_DOUBLE */
typedef signed int	INT;		/**< integer, at least 32 bits */
typedef unsigned int	UINT;		/**< unsigned integer */
typedef signed short	SHORT;		/**< short, at least 16 bits */
typedef unsigned short	USHORT;		/**< unsigned short */
typedef signed char	CHAR;		/**< character */
typedef unsigned char	BYTE;		/**< byte */

#if USE_MPI
# if USE_LONG_DOUBLE
#  define PHG_MPI_FLOAT	MPI_LONG_DOUBLE
# else	/* USE_LONG_DOUBLE */
#  define PHG_MPI_FLOAT	MPI_DOUBLE
# endif	/* USE_LONG_DOUBLE */
# define PHG_MPI_INT	MPI_INT
# define PHG_MPI_UINT	MPI_UNSIGNED
# define PHG_MPI_SHORT	MPI_SHORT
# define PHG_MPI_USHORT	MPI_UNSIGNED_SHORT
# define PHG_MPI_CHAR	MPI_CHAR
# define PHG_MPI_BYTE	MPI_BYTE
#endif	/* USE_MPI */

/* some mathematical functions, compatible with the FLOAT type */
#if USE_LONG_DOUBLE
# define FLOAT_MAX	LDBL_MAX
# define FLOAT_MIN	LDBL_MIN
# define FLOAT_EPSILON	LDBL_EPSILON
# define Pow/*(a,b)*/	powl/*((FLOAT)(a),(FLOAT)(b))*/
# define Sqrt/*(a)*/	sqrtl/*((FLOAT)(a))*/
# define Fabs/*(a)*/	fabsl/*((FLOAT)(a))*/
# define Log/*(a)*/	logl/*((FLOAT)(a))*/
# define Exp/*(a)*/	expl/*((FLOAT)(a))*/
# define Sin/*(a)*/	sinl/*((FLOAT)(a))*/
# define Asin/*(a)*/	asinl/*((FLOAT)(a))*/
# define Cos/*(a)*/	cosl/*((FLOAT)(a))*/
# define Acos/*(a)*/	acosl/*((FLOAT)(a))*/
# define Tan/*(a)*/	tanl/*((FLOAT)(a))*/
# define Atan/*(a)*/	atanl/*((FLOAT)(a))*/
# define Floor/*(a)*/	floorl/*((FLOAT)(a))*/
# define Ceil/*(a)*/	ceill/*((FLOAT)(a))*/
# else	/* USE_LONG_DOUBLE */
# define FLOAT_MAX	DBL_MAX
# define FLOAT_MIN	DBL_MIN
# define FLOAT_EPSILON	DBL_EPSILON
# define Pow/*(a,b)*/	pow/*((FLOAT)(a),(FLOAT)(b))*/
# define Sqrt/*(a)*/	sqrt/*((FLOAT)(a))*/
# define Fabs/*(a)*/	fabs/*((FLOAT)(a))*/
# define Log/*(a)*/	log/*((FLOAT)(a))*/
# define Exp/*(a)*/	exp/*((FLOAT)(a))*/
# define Sin/*(a)*/	sin/*((FLOAT)(a))*/
# define Asin/*(a)*/	asin/*((FLOAT)(a))*/
# define Cos/*(a)*/	cos/*((FLOAT)(a))*/
# define Acos/*(a)*/	acos/*((FLOAT)(a))*/
# define Tan/*(a)*/	tan/*((FLOAT)(a))*/
# define Atan/*(a)*/	atan/*((FLOAT)(a))*/
# define Floor/*(a)*/	floor/*((FLOAT)(a))*/
# define Ceil/*(a)*/	ceil/*((FLOAT)(a))*/
#endif	/* USE_LONG_DOUBLE */

#undef TRUE
#define TRUE	(1)
#undef FALSE
#define FALSE	(0)
typedef int BOOLEAN;			/**< boolean */

/* Geometric types. Also used for refinement types: EDGE means the refinement
   edge is an edge of the reference parallelepiped, FACE means the refinement
   edge is the diagonal of a face of the reference parallelepiped, and
   DIAGONAL means the refinement edge is the diagonal of the reference
   parallelepiped */
#define GTypeName(t) (					\
	((t) == VERTEX)		? "VERTEX" :		\
	((t) == EDGE)		? "EDGE" :		\
	((t) == FACE)		? "FACE" :		\
	((t) == DIAGONAL)	? "DIAGONAL" :		\
	((t) == OPPOSITE)	? "OPPOSITE" :		\
	((t) == MIXED)		? "MIXED" : "UNKNOWN")
enum {VERTEX, EDGE, FACE, DIAGONAL, OPPOSITE, MIXED};
typedef BYTE GTYPE;	/* force to BYTE to save memory */

#define ELEMENT		DIAGONAL
#define P_VERTEX	OPPOSITE

/* types of faces (REMOTE means it has a remote neighbour), the type
   UNREFERENCED means the object does not belong to the leaf elements.

   Note:
   1. multiple bits may be set, e.g., DIRICHLET | REMOTE | INTERIOR
   2. REMOTE bit set ==> INTERIOR bit set
*/
#define BTypeName(t) (				\
	((t) & UNREFERENCED)	? "unused" :	\
	((t) & INTERIOR)	? (((t) & REMOTE) ? "remote" : "interior") : \
	((t) & DIRICHLET)	? "Dirichlet":	\
	((t) & NEUMANN)		? "Neumann" :	\
	((t) & BDRY_USER1)	? "user 1" :	\
	((t) & BDRY_USER2)	? "user 2" :	\
	((t) & UNDEFINED)	? "undefined" : "invalid")
enum {
    UNREFERENCED= 0,	/**< not referenced by leaf elements of the submesh */
    OWNER	= 1,	/**< assigned to current submesh */
    INTERIOR    = 2,	/**< interior of the mesh */
    REMOTE	= 4,	/**< shared with a remote process */
    DIRICHLET   = 8,	/**< belongs to a Dirichlet boundary */
    NEUMANN	= 16,	/**< belongs to a Neumann boundary */
    BDRY_USER1	= 32,	/**< belongs to a user type 1 boundary */
    BDRY_USER2	= 64,	/**< belongs to a user type 2 boundary */
    UNDEFINED	= 128	/**< belongs to a boundary but has no bdry type */
};
typedef BYTE BTYPE;	/**< boundary type, force to BYTE to save memory */

#define BDRY_MASK (DIRICHLET | NEUMANN | BDRY_USER1 | BDRY_USER2 | UNDEFINED)
#define IS_BDRY(b)	((b) & BDRY_MASK)

/* GRID flags */
enum {
    VERT_FLAG = 1,
    EDGE_FLAG = 2,
    FACE_FLAG = 4,
    ELEM_FLAG = 8,
    GEOM_FLAG = 16
};

typedef FLOAT COORD[Dim];		/* space coordinates */

/**
 @brief The SIMPLEX struct
 */
typedef struct SIMPLEX_ {
    struct SIMPLEX_	*children[2];		/**< Pointers to children */
    void		*neighbours[NFace];	/**< Pointers to neighbours.
						   It stores the
						   index in g->neighbours.list
						   if the neighbour is remote */
#if 0
    void		*userdata;		/**< Pointer to userdata, for
						   use by applications */
#endif
    void		*parent;		/* pointer to parent */
    INT			verts[NVert];		/**< List of vertices */
    INT			edges[NEdge];		/**< List of edges */
#if (Dim == 3)
    INT			faces[NFace];		/**< List of faces */
#endif
    INT			index;			/**< Element index */
    SHORT		mark;			/**< Refine/coarsen mark */
    SHORT		region_mark;		/**< Regional mark */
#if ALLOW_CURVED_BOUNDARY
    /* Note: storing the indices vertex-wise instead of edge-wise saves
       memory, but it has problems when dealing with vertices across the
       interface of two surfaces */
    CHAR		bound_func[NEdge];	/**< Indices to curved boundary
						   edges (-1 if not curved) */
#endif
    BTYPE		bound_type[NFace];	/**< Boundary types */
    BYTE		ordering;		/**< Ordering of vertices:
						   - bits 0-1: 1st vertex
						   - bits 2-3: 2nd vertex
						   - bits 4-5: 3rd vertex
						   - bits 6-7: 4th vertex */
    CHAR		generation;		/**< Generation (level) */
    /*union {*/
	BYTE		flag;			/**< Scratch flag */
	CHAR		hp_order;		/**< for phgHPSetup() */
    /*};*/
#if (Dim == 3)
    GTYPE		type;			/**< Refinement type */
#endif
} SIMPLEX;

#if USE_MPI
# define GlobalVertex(g,no) \
	((g) == NULL || (g)->L2Gmap_vert==NULL ? (no):(g)->L2Gmap_vert[no])
# define GlobalEdge(g,no) \
	((g) == NULL || (g)->L2Gmap_edge==NULL ? (no):(g)->L2Gmap_edge[no])
# define GlobalFace(g,no) \
	((g) == NULL || (g)->L2Gmap_face==NULL ? (no):(g)->L2Gmap_face[no])
# define GlobalElement(g,no) \
	((g) == NULL || (g)->L2Gmap_elem==NULL ? (no):(g)->L2Gmap_elem[no])
#else
# define GlobalVertex(g,no) (no)
# define GlobalEdge(g,no) (no)
# define GlobalFace(g,no) (no)
# define GlobalElement(g,no) (no)
#endif

#define GlobalVertexP(g, no) \
    ((g)->period == NULL /*|| (g)->period->L2Gmap_vert == NULL*/ ? \
	GlobalVertex(g, no) : (g)->period->L2Gmap_vert[no])

#define ForAllElements(g, e) \
    for (_phg_i = 0; (e = _phg_step_element(g)) != NULL; _phg_i++)

#if USE_MPI
/* information about a remote neighbour of an element */
typedef struct {
    SIMPLEX	*remote;	/* addr of the neighbour on remote process */
    SIMPLEX	*local;		/* local element */
#if USE_ANONYMOUS_UNION
    union {
#endif	/* USE_ANONYMOUS_UNION */
	UINT	path;		/* bitwise flags recording which child to
				   select when refining a boundary edge.
				   0=child 0, 1=child 1 (used by refine.c) */
	INT	peer_face;	/* local index of the shared face in the peer
				   process (set by phgUpdateBoundaryTypes and
				   used by build_L2Vmap) */
#if USE_ANONYMOUS_UNION
    };
#endif	/* USE_ANONYMOUS_UNION */
    SHORT	rank;		/* remote process rank */
    BYTE	rface[3];	/* the matching face of remote neighbour */
    BYTE	vertex;		/* the face shared with remote process */
    BYTE	op_vertex;	/* opposite vertex on remote process */
    BYTE	depth;		/* refinement count (depth) of the element */
} RNEIGHBOUR;
#endif

/**
 @brief The GRID struct
 */
#define nleaf_global	nelem_global
typedef struct GRID_ {
    FLOAT	lif;		/**< Load imbalance factor */
    FLOAT	bbox[2][Dim];	/**< BoundingBox of the mesh */
    FLOAT	volume;		/**< Volume of the mesh */
    char	*filename;	/**< name of the mesh file */
    COORD	*verts;		/**< List of vertices */
    SIMPLEX	*roots;		/**< Root elements */
    SIMPLEX	**elems;	/**< Array of size g->nlem, pointers to leaf
				     elements (for mesh traversal), NULL if
				     an element index is unreferenced */
    struct DOF_	*geom;		/**< DOF containing the geometric data,
				     maintained by geom.c */
    struct DOF_	**dof;		/**< NULL terminated list of DOFs
				     defined on the mesh */
    struct HP_TYPE_ **hp;	/**< NULL terminated list of HP_TYPEs
				     defined on the mesh */

#if ALLOW_CURVED_BOUNDARY
    EXPR	**bdry_funcs;	/**< NULL terminated list of projection
				   functions for curved boundaries */
#endif
    BYTE	*types_vert;	/**< Types of vertices (bit flags) */
    BYTE	*types_edge;	/**< Types of edges (bit flags) */
    BYTE	*types_face;	/**< Types of faces (bit flags) */
    BYTE	*types_elem;	/**< Types of elements (bit flags) */

    /* inter-grid links */
    struct GRID_ *alien;	/**< Pointer to another grid */
    INT		*alien_map;	/**< Map of elements between the two grids */

    /* struct for handling periodic BC */
    struct PERIOD_ *period;

    /* Arrays containing ranks of, and local indices in the owner process
     * for vertices. They are non nil only when needed and are updated by
     * phgUpdateBoundaryTypes in utils.c and used by build_L2Vmap() in map.c.
     *
     * Note: an entry has value -1 iff it's uniquely owned. */
    INT		*owner_index_vert;
    SHORT	*owner_rank_vert;

#if USE_MPI
    struct {
	/* elements with neighbours on other processes are stored here */
	int		*counts;	/* Counts of remote faces */
	int		*displs;	/* Start positions in list */
	RNEIGHBOUR	*list;		/* List of remote faces */
	INT		count;		/* Total number of remote faces */
	INT		allocated;	/* Allocated size of list */
    }		neighbours;	/**< List of remote faces */
    /* Note: don't access L2Gmaps directly, use the GlobalXxxx macros instead */
    INT		*L2Gmap_vert;	/**< Local to global map of vertices */
    INT		*L2Gmap_edge;	/**< Local to global map of edges */
    INT		*L2Gmap_face;	/**< Local to global map of faces */
    INT		*L2Gmap_elem;	/**< Local to global map of element indices */

    /* Arrays containing ranks of, and local indices in the owner process
     * for edges, faces and elements. They are non nil only when
     * needed and are updated by phgUpdateBoundaryTypes in utils.c
     * and used by build_L2Vmap() in map.c.
     *
     * Note: an entry has value -1 iff it's uniquely owned. */
    INT		*owner_index_edge;
    SHORT	*owner_rank_edge;

    /* Note: owner_xxxx_face and owner_xxxx_elem are currently not used */
    INT		*owner_index_face;
    SHORT	*owner_rank_face;

    INT		*owner_index_elem;
    SHORT	*owner_rank_elem;

    MPI_Comm	comm;		/**< The mesh's communicator */
#else	/* USE_MPI */
    int		comm;
#endif	/* USE_MPI */
    INT		nleaf;		/**< Number of leaf elements in the submesh */
    INT		nvert;		/**< Number of vertices in the submesh */
    INT		nedge;		/**< Number of edges in the submesh */
#if Dim == 3
    INT		nface;		/**< Number of faces in the submesh */
#endif
    INT		nelem;		/**< Number of element indices in the submesh */

    /* counts of owned vertices/edges/faces/elements, they are updated by
     * the function phgUpdateBoundaryTypes */
    INT		nvert_owned;
    INT		nedge_owned;
    INT		nface_owned;
    INT		nelem_owned;

    /* global counters. 2D, V:E:T \approx 1:3:2. 3D, V:E:F:T \approx 1:6:10:5 */
    INT		nvert_global;	/**< Number of vertices in the global mesh */
    INT		nedge_global;	/**< Number of edges in the global mesh */
#if Dim == 3
    INT		nface_global;	/**< Number of faces in the global mesh */
#endif
    INT		nelem_global;	/**< Number of tetrahedra in the global mesh */

    INT		nroot;		/**< Number of root elements */
    INT		ntree;		/**< Total number of elements in the tree */
    INT		serial_no;	/**< Number of grid changes */
    int		last_partitioner; /**< Last partitioner used */
    int		rank;		/**< Process rank (submesh no) */
    int		nprocs;		/**< Number of processes for the mesh */
    int		bc_rmap[5];	/**< Saves PHG ==> mesh file bc map */
    BYTE	flags;		/** {VERT,EDGE,FACE,ELEM,GEOM}_FLAG bits */
} GRID;

#include <float.h>

#define IsZero(a) (Fabs(a) < 10.0 * FLOAT_MIN)
#define IsOne(a)  (Fabs(a - 1.0) < 10.0 * FLOAT_EPSILON)

#include "utils.h"
#include "refine.h"
#include "coarsen.h"
#include "phg-mpi.h"
#include "albert.h"
#if USE_MPI
# include "mpi-utils.h"
#endif
#include "medit.h"
#include "dof.h"

#include "distribute.h"

#include "lagrange.h"
#include "nedelec.h"
#include "hfeb.h"
#include "hierarchical-basis.h"
#include "dg.h"

#include "quad.h"
#include "quad-gauss.h"

#include "geom.h"

#include "matvec.h"
#include "map.h"
#include "solver.h"

#include "vtk.h"
#include "opendx.h"

#include "option.h"
#include "perf.h"
#include "eigen.h"
#include "dof-utils.h"
#include "mark.h"

#include "surface-cut.h"
#include "intergrid.h"

#include "partition-utils.h"
#include "hp.h"

/* Note: only need to export function phgSetPeriodicity to users */
/*#include "periodic.h"*/
#define X_MASK	1
#define Y_MASK	2
#define Z_MASK	4

#ifdef __cplusplus
extern "C" {
#endif

void phgSetPeriodicity(GRID *g, int flags);
void phgSetPeriodicDirections(GRID *g, const FLOAT dirs[]);

#ifdef __cplusplus
}
#endif

#define PHG_PHG_H
#endif
